==> inca_tool-1.0.0-py3-none-any.whl: inca/woodprovision/preprocessing/dmp.py <==
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.

"""Preprocess DMP from the PROBAV or SENTINEL-3 archive.

The total dry matter productivity of the year is calculated out of all input data in a folder. Then the raster is
resampled plus warped to the needed AOI and resolution.

Important: the 10daily input DMP data must be geotiffs (see lower comment for conversion).

PREREQUISITES

If you do not have 10-daily DMP data then you have to extract it from the CGLS:
- I use the cgls-geotiff package from VITO to access the right data and transform from netcdf to geotiff
1. clone the repository --> https://git.vito.be/scm/biggeo/cgls-geotiff.git
2. import it in folder on the MEP (e.g. ~/workspace/cgls_geotiff )
3. go to the root folder and do an editable install: python3 -m pip install --user -e .
4. check out the branch "dmp-inca".
5. now you can call in the terminal the processing of the nc files, e.g.

    python3 DMP_preparation_INCA.py BioPar_DMP_V2_Global 20000101 20001231
        /data/inca_vol1/input_data/CGL_BioPar_DMP_V2_Global True

"""
import argparse
import logging
import os
import sys
import time
import subprocess
from datetime import datetime, timedelta

import rasterio
import numpy as np

import inca
import inca.common.geoprocessing as geo
import inca.woodprovision.preprocessing

logger = logging.getLogger(__name__)

description = 'DMP preprocessing for wood provision.'


def main():
    start_time = time.time()
    options = parse_args()

    inca.set_up_console_logging(options.verbose)

    logger.info('START OF MODULE: %s', description)

    # init the metadata and raster extent of project
    metadata = geo.Metadata('INCA Tool', description)

    os.makedirs(options.path_out, exist_ok=True)

    logger.info('Generate the total dmp in original resolution')
    try:
        path_dmp = os.path.join(options.path_out, f'DMP_{options.year}_annual-total_INCA.tif')
        scale_factor = processing_total_dmp(options, metadata, path_dmp)
    except Exception as e:
        if os.path.exists(path_dmp):
            os.remove(path_dmp)
        logger.error('Failed to generate total annual DMP for year %s: "%s".', options.year, e)
        sys.exit(1)

    logger.info('Bring the final DMP product to AOI')
    try:
        bring_2_AOI(options, metadata, path_dmp, scale_factor)
    except Exception as e:
        logger.error('Failed to bring yearly DMP  to reference AOI: "%s"', e)
        sys.exit(1)

    logger.info('All processing successfully done. Have a nice day.')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: %s', description)


def parse_args():
    parser = argparse.ArgumentParser(description)
    parser.add_argument('path_reference', help='Path to reference raster file for projection/resolution/extent.',
                        metavar='reference_raster')
    parser.add_argument('path_in', help='Path to folder with DMP input data.', metavar='dmp_dir')
    parser.add_argument('path_out', help='Path to output folder.', metavar='output_dir')
    parser.add_argument('year', type=int, help='Processing year.')
    parser.add_argument('-v', '--verbose', action='store_true', help='verbose output')

    return parser.parse_args()


def bring_2_AOI(options, metadata, path_dmp, scale_factor):
    """Warp DMP original dataset to correct EPSG, resolution and AOI."""
    path_out = os.path.join(os.path.normpath(options.path_out), 'DMP_{}_annual_total_INCA_{}m_EPSG{}.tif'.format(
        options.year,
        int(metadata.profile['transform'].a),
        metadata.profile['crs'].to_epsg()))

    cmd = 'gdalwarp --config GDAL_CACHEMAX 256 -t_srs {} -te {} {} {} {} -tr {} {} -r {} -et 0 -nomd ' \
        '-wo SAMPLE_STEPS=50 -wo SOURCE_EXTRA=5 -wo SAMPLE-GRID=YES '\
        '-co COMPRESS=LZW -co TILED=YES -co INTERLEAVE=BAND -co BIGTIFF=YES -multi -overwrite {} {}'.format(
            metadata.profile['crs'].to_string(),
            metadata.pextent.left,
            metadata.pextent.bottom,
            metadata.pextent.right,
            metadata.pextent.top,
            metadata.profile['transform'].a,
            abs(metadata.profile['transform'].e),
            'average',
            path_dmp,
            path_out)
    subprocess.check_call(cmd, shell=True)
    # extra metadata
    metadata.read_raster_tags([path_dmp])
    with rasterio.open(path_out, 'r+') as dst:
        metadata.update_dataset_tags(dst, f'Annual total of DMP derived from CGLS datasets for year {options.year}. '
                                     f'value_range = [0, 250]. scaling_factor: {scale_factor}. '
                                     'Warped to EPSG, resolution and extent of INCA AOI',
                                     'DMP', dst.name)


def processing_total_dmp(options, metadata, path_out, scale_factor=0.01, valid_range=[0, 32767]):
    """Calculate the total dmp out of given files in folder.

    NOTE: No check are performed - all tif files in folder are used to calculate total.

    standard for CGLS 1km product
    valid_range: 0 - 32767
    nodata:  -1,-2
    scaling: 0.001
    """
    # we do not care about memory so we run all in one chunk
    # get list of all files in input folder
    lFiles = [entry.path for entry in os.scandir(os.path.normpath(options.path_in))
              if entry.is_file() and entry.name.endswith(".tiff")]

    # get main profile data from first file
    with rasterio.open(lFiles[0]) as src:
        src_profile = src.profile
        try:
            s = float(src.tags(1)['scale_factor'])
            if (s >= 0) and (s <= 1.):
                scale_factor = s
        except Exception:
            logger.info('use standard scale factor since no matching metadata.')
        try:
            v = [int(x) for x in (src.tags(1)['valid_range']).replace(']', ' ').split(' ') if x.isnumeric()]
            if (len(v) == 2) and (v[0] < v[1]):
                valid_range = v
        except Exception:
            logger.info('use standard valid data range since no matching metadata.')

    # ini the arrays for sum and count
    aSum = np.zeros((src_profile['height'], src_profile['width']), dtype=np.uint)
    aDays = np.zeros_like(aSum, dtype=np.uint)
    maxDays = 0

    # loop over all files
    for element in lFiles:
        with rasterio.open(element) as src:
            aData = src.read(1).astype(np.uint)

            time_coverage_start = datetime.strptime(src.tags()['time_coverage_start'],  '%Y-%m-%dT%H:%M:%SZ')
            time_coverage_end = datetime.strptime(src.tags()['time_coverage_end'],  '%Y-%m-%dT%H:%M:%SZ')

        # get valid data
        mValid = (aData >= valid_range[0]) & (aData <= valid_range[1])

        # fill prepared arrays
        aSum[mValid] += aData[mValid]*(time_coverage_end-time_coverage_start + timedelta(seconds=1)).days
        aDays[mValid] += (time_coverage_end-time_coverage_start + timedelta(seconds=1)).days
        maxDays += (time_coverage_end-time_coverage_start + timedelta(seconds=1)).days

    # correcting for missing data with yearly average
    # aSum[aDays < maxDays] = aSum[aDays < maxDays]/aDays[aDays < maxDays]*maxDays

    # setup dst profile
    src_profile.update(count=1, dtype=rasterio.int32, nodata=-1)

    # output filename
    with rasterio.open(path_out, 'w', **src_profile) as dst:
        metadata.update_dataset_tags(dst,
                                     f'Annual total of DMP derived from CGLS datasets for the year {options.year}. '
                                     f'value_range = [0, 11959955]. scaling_factor: {scale_factor}',
                                     'DMP', *lFiles)
        dst.update_tags(1, scale_factor=scale_factor)
        dst.write(aSum.astype(src_profile['dtype']), 1)
    return scale_factor


if __name__ == '__main__':
    # hack to get the proper logger when running the script directly
    logger = logging.getLogger(inca.woodprovision.preprocessing.__name__)
    main()

==> inca_tool-1.1.2-py3-none-any.whl: inca/woodprovision/preprocessing/dmp.py <==
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.

"""Preprocess DMP from the PROBAV or SENTINEL-3 archive.

The total dry matter productivity of the year is calculated out of all input data in a folder. Then the raster is
resampled plus warped to the needed AOI and resolution.

Important: the 10daily input DMP data must be geotiffs (see lower comment for conversion).

PREREQUISITES

If you do not have 10-daily DMP data then you have to extract it from the CGLS:
- I use the cgls-geotiff package from VITO to access the right data and transform from netcdf to geotiff
1. clone the repository --> https://git.vito.be/scm/biggeo/cgls-geotiff.git
2. import it in folder on the MEP (e.g. ~/workspace/cgls_geotiff )
3. go to the root folder and do an editable install: python3 -m pip install --user -e .
4. check out the branch "dmp-inca".
5. now you can call in the terminal the processing of the nc files, e.g.

    python3 DMP_preparation_INCA.py BioPar_DMP_V2_Global 20000101 20001231
        /data/inca_vol1/input_data/CGL_BioPar_DMP_V2_Global True

"""
import argparse
import logging
import os
import sys
import time
import subprocess
from datetime import datetime, timedelta

import rasterio
import numpy as np

import inca
import inca.common.geoprocessing as geo
import inca.woodprovision.preprocessing

logger = logging.getLogger(__name__)

description = 'DMP preprocessing for wood provision.'


def main():
    start_time = time.time()
    options = parse_args()

    inca.set_up_console_logging(options.verbose)

    logger.info('START OF MODULE: %s', description)

    # init the metadata and raster extent of project
    metadata = geo.Metadata('INCA Tool', description)

    os.makedirs(options.path_out, exist_ok=True)

    logger.info('Generate the total dmp in original resolution')
    try:
        path_dmp = os.path.join(options.path_out, f'DMP_{options.year}_annual-total_INCA.tif')
        scale_factor = processing_total_dmp(options, metadata, path_dmp)
    except Exception as e:
        if os.path.exists(path_dmp):
            os.remove(path_dmp)
        logger.error('Failed to generate total annual DMP for year %s: "%s".', options.year, e)
        sys.exit(1)

    logger.info('Bring the final DMP product to AOI')
    try:
        pass
        bring_2_AOI(options, metadata, path_dmp, scale_factor)
    except Exception as e:
        logger.error('Failed to bring yearly DMP  to reference AOI: "%s"', e)
        sys.exit(1)

    logger.info('All processing successfully done. Have a nice day.')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: %s', description)


def parse_args():
    parser = argparse.ArgumentParser(description)
    parser.add_argument('path_reference', help='Path to reference raster file for projection/resolution/extent.',
                        metavar='reference_raster')
    parser.add_argument('path_in', help='Path to folder with DMP input data.', metavar='dmp_dir')
    parser.add_argument('path_out', help='Path to output folder.', metavar='output_dir')
    parser.add_argument('year', type=int, help='Processing year.')
    parser.add_argument('-v', '--verbose', action='store_true', help='verbose output')

    return parser.parse_args()


def bring_2_AOI(options, metadata, path_dmp, scale_factor):
    """Warp DMP original dataset to correct EPSG, resolution and AOI."""
    path_out = os.path.join(os.path.normpath(options.path_out), 'DMP_{}_annual_total_INCA_{}m_EPSG{}.tif'.format(
        options.year,
        int(metadata.profile['transform'].a),
        metadata.profile['crs'].to_epsg()))

    cmd = 'gdalwarp --config GDAL_CACHEMAX 256 -t_srs {} -te {} {} {} {} -tr {} {} -r {} -et 0 -nomd ' \
        '-wo SAMPLE_STEPS=50 -wo SOURCE_EXTRA=5 -wo SAMPLE-GRID=YES '\
        '-co COMPRESS=LZW -co TILED=YES -co INTERLEAVE=BAND -co BIGTIFF=YES -multi -overwrite {} {}'.format(
            metadata.profile['crs'].to_string(),
            metadata.pextent.left,
            metadata.pextent.bottom,
            metadata.pextent.right,
            metadata.pextent.top,
            metadata.profile['transform'].a,
            abs(metadata.profile['transform'].e),
            'average',
            path_dmp,
            path_out)
    subprocess.check_call(cmd, shell=True)
    # extra metadata
    metadata.read_raster_tags([path_dmp])
    with rasterio.open(path_out, 'r+') as dst:
        metadata.update_dataset_tags(dst, f'Annual total of DMP derived from CGLS datasets for year {options.year}. '
                                     f'value_range = [0, 250]. scaling_factor: {scale_factor}. '
                                     'Warped to EPSG, resolution and extent of INCA AOI',
                                     'DMP', dst.name)


def processing_total_dmp(options, metadata, path_out, scale_factor=0.01, valid_range=[0, 32767]):
    """Calculate the total dmp out of given files in folder.

    NOTE: No check are performed - all tif files in folder are used to calculate total.

    standard for CGLS 1km product
    valid_range: 0 - 32767
    nodata:  -1,-2
    scaling: 0.001
    """
    # we do not care about memory so we run all in one chunk
    # get list of all files in input folder
    lFiles = [entry.path for entry in os.scandir(os.path.normpath(options.path_in))
              if entry.is_file() and entry.name.endswith(".tiff")]

    # get main profile data from first file
    with rasterio.open(lFiles[0]) as src:
        src_profile = src.profile
        try:
            s = float(src.tags(1)['scale_factor'])
            if (s >= 0) and (s <= 1.):
                scale_factor = s
        except Exception:
            logger.info('use standard scale factor since no matching metadata.')
        try:
            v = [int(x) for x in (src.tags(1)['valid_range']).replace(']', ' ').split(' ') if x.isnumeric()]
            if (len(v) == 2) and (v[0] < v[1]):
                valid_range = v
        except Exception:
            logger.info('use standard valid data range since no matching metadata.')


        # output filename
    # setup dst profile
    src_profile.update(count=1, dtype=rasterio.int32, nodata=-1)
    with rasterio.open(path_out, 'w', **src_profile) as dst:
        metadata.update_dataset_tags(dst,
                                     f'Annual total of DMP derived from CGLS datasets for the year {options.year}. '
                                     f'value_range = [0, 11959955]. scaling_factor: {scale_factor}',
                                     'DMP', *lFiles)
        dst.update_tags(1, scale_factor=scale_factor)


        # loop over all files

        for _, window in geo.block_window_generator((2048,2048), dst.height, dst.width):
            for idx,element in enumerate(lFiles):
                with rasterio.open(element) as src:
                    aData = src.read(1, window=window).astype(np.uint)
                    if idx == 0:
                        # ini the arrays for sum and count
                        aSum = np.zeros_like(aData, dtype=np.uint)
                        aDays = np.zeros_like(aData, dtype=np.uint)
                        maxDays = 0

                    time_coverage_start = datetime.strptime(src.tags()['time_coverage_start'],  '%Y-%m-%dT%H:%M:%SZ')
                    time_coverage_end = datetime.strptime(src.tags()['time_coverage_end'],  '%Y-%m-%dT%H:%M:%SZ')

                # get valid data
                mValid = (aData >= valid_range[0]) & (aData <= valid_range[1])

                # fill prepared arrays
                aSum[mValid] += aData[mValid]*(time_coverage_end-time_coverage_start + timedelta(seconds=1)).days
                aDays[mValid] += (time_coverage_end-time_coverage_start + timedelta(seconds=1)).days
                maxDays += (time_coverage_end-time_coverage_start + timedelta(seconds=1)).days

            # correcting for missing data with yearly average
            # aSum[aDays < maxDays] = aSum[aDays < maxDays]/aDays[aDays < maxDays]*maxDays

            dst.write(aSum.astype(src_profile['dtype']), 1, window= window)
    return scale_factor


if __name__ == '__main__':
    # hack to get the proper logger when running the script directly
    logger = logging.getLogger(inca.woodprovision.preprocessing.__name__)
    main()
